// These are the general routines for supporting MFM decoders.
// Call mfm_decode_setup once before trying to decode a disk
// call mfm_init_sector_status_list before decoding each track
// call mfm_update_deltas before decoding each track
// call mfm_decode_track to decode the track or
//   call mfm_wait_all_deltas to just wait for all deltas to be received and
//   possibly written to a file
// call mfm_check_deltas to get number of deltas in the buffer
// call mfm_check_header_values to verify header parameters
// call mfm_write_sector to check sector information and possibly write it to
//   the extract file
// call mfm_decode_done when all tracks have been processed
//
// TODO: make it use sector number information and checking CRC at data length to write data
// for sectors with bad headers. See if resyncing PLL at write boundaries improves performance when
// data bits are shifted at write boundaries.

// Copyright 2014 David Gesswein.
// This file is part of MFM disk utilities.
//
// MFM disk utilities is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// MFM disk utilities is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with MFM disk utilities.  If not, see <http://www.gnu.org/licenses/>.
#include <stdio.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <stdint.h>
#include <limits.h>

#include "msg.h"
#include "crc_ecc.h"
#include "mfm_decoder.h"

#define ARRAYSIZE(x)  (sizeof(x) / sizeof(x[0]))

// Hold for sector status and cylinder and head it was for. We save the
// data so when the cylinder and head changes we can print the final status.
// The same track may be reread.
static SECTOR_STATUS last_sector_list[MAX_SECTORS];
static int last_cyl;
static int last_head;
// File to write decoded data to. -1 if shoudln't write to file.
static int decode_fd = -1;

// These are various PLL constants I was trying. For efficiency the
// filter routine is put in the decoders so it can inline and has the
// "best" PLL constants I found.
#if 0
int filter_type;

static inline float filter(float v, float *delay)
{
   float in, out;

   in = v + *delay;
#if 1
   out = in * 0.034446428576716f + *delay * -0.034124999994713f;
#else
   switch (filter_type) {
   case 0:
      out = in * 0.034446428576716f + *delay * -0.034124999994713f;
      break;
   case 1:
      out = in * 0.114928571449721f + *delay * -0.113642857121708f;     // 1, 2e-5
      break;
   case 2:
      out = in * 0.034928571449721f + *delay * -0.033642857121708f;     // .3, 2e-5
      break;
   case 3:
      out = in * 0.071142857227454f + *delay * -0.065999999915403f;     // .3, 1e-5
      break;
   case 4:
      o1t = in * 0.023114285722745f + *delay * -0.022599999991540f;     // .3, 10e-5, /10
      break;
   case 5:
      out = in * 0.068828571437031f + *delay * -0.068314285705826f;     // .3, 10e-5, /10
      break;
   }
#endif

   *delay = in;
   return out;
}
#endif

// Print any errors and ECC correction from the sector_status_list.
//
// sector_status_list: List of sector statuses
// num_sectors: Length of list
// cyl, head: Track status if for
static void print_sector_list_status(SECTOR_STATUS *sector_status_list,
      int num_sectors, int cyl, int head) {
   int cntr;
   int ecc_corrections = 0;
   int hard_errors = 0;

   // Find if anything needs printing
   for (cntr = 0; cntr < num_sectors; cntr++) {
      if (UNRECOVERED_ERROR(sector_status_list[cntr].status)) {
         hard_errors = 1;
      }
      if (sector_status_list[cntr].status & SECT_ECC_RECOVERED) {
         ecc_corrections = 1;
      }
   }
   // And if so print it
   if (hard_errors) {
      msg(MSG_ERR_SUMMARY, "Bad sectors on cylinder %d head %d:",cyl,head);
      for (cntr = 0; cntr < num_sectors; cntr++) {
         if (sector_status_list[cntr].status & SECT_BAD_HEADER) {
            msg(MSG_ERR_SUMMARY, " %dH", cntr);
         }
         if (sector_status_list[cntr].status & SECT_BAD_DATA) {
            msg(MSG_ERR_SUMMARY, " %d", cntr);
         }
      }
      msg(MSG_ERR_SUMMARY,"\n");
   }
   if (ecc_corrections) {
      msg(MSG_ERR_SUMMARY, "ECC Corrections on cylinder %d head %d:",cyl,head);
      for (cntr = 0; cntr < num_sectors; cntr++) {
         if (sector_status_list[cntr].status & SECT_ECC_RECOVERED) {
            msg(MSG_ERR_SUMMARY, " %d(", cntr);
            if (sector_status_list[cntr].ecc_span_corrected_header) {
               msg(MSG_ERR_SUMMARY, "%dH%s",
                     sector_status_list[cntr].ecc_span_corrected_header,
                     sector_status_list[cntr].ecc_span_corrected_data != 0 ? "," : "");
            }
            if (sector_status_list[cntr].ecc_span_corrected_data) {
               msg(MSG_ERR_SUMMARY, "%d", sector_status_list[cntr].ecc_span_corrected_data);
            }
            msg(MSG_ERR_SUMMARY,")");
         }
      }
      msg(MSG_ERR_SUMMARY, "\n");
   }
}

// Update statistics for the read so we can print summary at the end.
// Since we may be called multiple times for the same track if errors are
// retried only update the statistics and print errors when the track
// changes.
//
// drive_params: Parameters for drive. Stats updated.
// cyl, head: cylinder and head from header
// sector_status_list: status of sectors
static void update_stats(DRIVE_PARAMS * drive_params, int cyl, int head,
      SECTOR_STATUS sector_status_list[])
{
   STATS *stats = &drive_params->stats;
   int i;
   int error_found = 0;

   // If track changed and list has been set (last_cyl != -1) then process
   if (last_cyl != -1 && (cyl != last_cyl || head != last_head)) {
      for (i = 0; i < drive_params->num_sectors; i++) {
         if (last_sector_list[i].status & SECT_ECC_RECOVERED) {
            stats->num_ecc_recovered++;
            error_found = 1;
         }
         if (last_sector_list[i].status & SECT_BAD_DATA) {
            stats->num_bad_data++;
            error_found = 1;
         } else if (last_sector_list[i].status & SECT_BAD_HEADER) {
            stats->num_bad_header++;
            error_found = 1;
         } else {
            stats->num_good_sectors++;
         }
      }
      if (error_found) {
         print_sector_list_status(last_sector_list, drive_params->num_sectors, last_cyl,
               last_head);
      }
   }
   // Save the sector information so we can use it when track changes
   if (sector_status_list != NULL) {
      memcpy(last_sector_list, sector_status_list, sizeof(last_sector_list));
   }
   last_cyl = cyl;
   last_head = head;
}

// Sets the sector status list to bad header. This is the default error.
// If we don't find the header when trying to read so can't update the list
// we will then have bad header status
//
// sector_status_list: status of sectors
// num_sectors: number of sectors in list
void mfm_init_sector_status_list(SECTOR_STATUS sector_status_list[], int num_sectors)
{
   int i;

   memset(sector_status_list, 0, sizeof(*sector_status_list) * num_sectors);
   for (i = 0; i < num_sectors; i++) {
      sector_status_list[i].status = SECT_BAD_HEADER;
   }
}

// This routine calls the proper decoder for the drive format
// See called routines for parameters and return value
SECTOR_DECODE_STATUS mfm_decode_track(DRIVE_PARAMS * drive_params, int cyl, int head,
      uint16_t deltas[], int *seek_difference,
      SECTOR_STATUS sector_status_list[])
{
   int rc;

   // Change in mfm_process_bytes if this if is changed
   if (drive_params->controller == CONTROLLER_WD_1006 ||
         drive_params->controller == CONTROLLER_OMTI_5510 ||
         drive_params->controller == CONTROLLER_DEC_RQDX3) {
      rc = wd_decode_track(drive_params, cyl, head, deltas, seek_difference,
            sector_status_list);
   } else {
      rc = xebec_decode_track(drive_params, cyl, head, deltas, seek_difference,
            sector_status_list);
   }
   update_stats(drive_params, cyl, head, sector_status_list);
   return rc;
}

// This makes the floating point faster with minor simplifications to the floating point processing
// http://infocenter.arm.com/help/index.jsp?topic=/com.arm.doc.ddi0344k/Chdiihcd.html
#ifdef __arm__
static void enable_runfast()
{
   static const unsigned int x = 0x04086060;
   static const unsigned int y = 0x03000000;
   int r;
   asm volatile ("fmrx %0, fpscr         \n\t"  //r0 = FPSCR
         "and  %0, %0, %1        \n\t"  //r0 = r0 & 0x04086060
         "orr  %0, %0, %2        \n\t"  //r0 = r0 | 0x03000000
         "fmxr fpscr, %0         \n\t"  //FPSCR = r0
         :"=r" (r)
          :"r"(x), "r"(y)
   );
}
#endif

// Setup to start decoding. Pass null pointer for filename if decoded data
// shouldn't be written to a file.
//
// filename: Filename to write decoded data to
// drive_params: Parameters for drive
void mfm_decode_setup(char *filename, DRIVE_PARAMS *drive_params)
{
   STATS *stats = &drive_params->stats;

   // set to value indicating not yet set.
   last_head = -1;
   last_cyl = -1;

   if (filename == NULL ) {
      decode_fd = -1;
   } else {
      decode_fd = open(filename, O_WRONLY | O_CREAT | O_TRUNC, 0664);
      if (decode_fd < 0) {
         perror("Unable to create output file");
         exit(1);
      }
   }
   memset(stats, 0, sizeof(*stats));
   stats->min_sect = INT_MAX;
   stats->min_head = INT_MAX;
   stats->min_cyl = INT_MAX;

#if __arm__
   // Speed up floating point on beaglebone
   enable_runfast();
#endif
}

// This cleans up at end of processing data and prints summary information.
// Also checks that data agrees with the drive parameters.
//
// drive_params: Parameters for drive
void mfm_decode_done(DRIVE_PARAMS * drive_params)
{
   STATS *stats = &drive_params->stats;

   // Process last track sector list
   update_stats(drive_params, -1, -1, NULL);
   if (decode_fd >= 0) {
      close(decode_fd);

      msg(MSG_STATS,
            "Found cyl %d to %d, head %d to %d, sector %d to %d\n",
            stats->min_cyl, stats->max_cyl, stats->min_head, stats->max_head,
            stats->min_sect, stats->max_sect);
      if (stats->max_cyl - stats->min_cyl + 1 != drive_params->num_cyl) {
         msg(MSG_ERR_SUMMARY, "Expected cyls %d doesn't match cyls found %d\n",
               drive_params->num_cyl, stats->max_cyl - stats->min_cyl + 1);
      }
      if (stats->max_head - stats->min_head + 1 != drive_params->num_head) {
         msg(MSG_ERR_SUMMARY, "Expected heads %d doesn't match heads found %d\n",
               drive_params->num_head, stats->max_head - stats->min_head + 1);
      }
      if (stats->max_sect - stats->min_sect + 1 != drive_params->num_sectors) {
         msg(MSG_ERR_SUMMARY, "Expected sectors %d doesn't match sectors found %d\n",
               drive_params->num_sectors, stats->max_sect - stats->min_sect + 1);
      }
      if (stats->min_sect != drive_params->first_sector_number) {
         msg(MSG_ERR_SUMMARY, "Expected first sector number %d doesn't match first sector found %d\n",
               drive_params->first_sector_number, stats->min_sect);

      }

      msg(MSG_STATS,
            "Expected %d sectors got %d good sectors, %d bad header, %d bad data\n",
            drive_params->num_cyl * drive_params->num_head *
            drive_params->num_sectors, stats->num_good_sectors,
            stats->num_bad_header, stats->num_bad_data);
      msg(MSG_STATS,
            "%d sectors corrected with ECC. Max bits in burst corrected %d\n",
            stats->num_ecc_recovered, stats->max_ecc_span);
   }
}

// This checks that the sector header values are reasonable and match the
// track and head we thought we were reading. If the cylinder doesn't match
// we return the difference between the actual and expected cylinder so
// the caller can try seeking again if needed.
//
// exp_cyl, exp_head: Track data was from
// sector_index: Counter for sectors starting at 0. With errors may not match
//   sector number
//  sector_size: Sector size from header
// seek_difference: Return of cylinder difference
// sector_status: Status of this sector
// drive_params: Parameters for drive
void mfm_check_header_values(int exp_cyl, int exp_head,
      int *sector_index, int sector_size, int *seek_difference,
      SECTOR_STATUS *sector_status, DRIVE_PARAMS *drive_params) {

   if (drive_params->ignore_header_mismatch) {
      sector_status->logical_sector = *sector_index;
      (*sector_index)++;
      return;
   }

   // WD 1003 controllers wrote 3 bit head code so head 8 is written as 0.
   // If requested and head seems correct fix read head value.
   if (drive_params->head_3bit && sector_status->head == (exp_head & 0x7)) {
      sector_status->head = exp_head;
   }
   if (sector_status->head != exp_head || sector_status->cyl != exp_cyl) {
      msg(MSG_ERR,"Mismatch cyl %d,%d head %d,%d index %d\n",
            sector_status->cyl, exp_cyl, sector_status->head, exp_head,
            *sector_index);
      sector_status->status |= SECT_BAD_HEADER;
      // Possibly a seek error, mark it
      if (sector_status->cyl != exp_cyl) {
         sector_status->status |= SECT_WRONG_CYL;
         if (seek_difference != NULL) {
            *seek_difference = exp_cyl - sector_status->cyl;
         }
      }
   }
   // If we have expected sector ordering information check the sector numbers
   // TODO: make this handle more complex sector numbering where they vary
   // between tracks
   if (drive_params->sector_numbers != NULL) {
      int orig_sector_index = *sector_index;
      for (; *sector_index < drive_params->num_sectors; (*sector_index)++) {
         if (sector_status->sector == drive_params->sector_numbers[*sector_index]) {
            break;
         }
      }
      if (*sector_index > orig_sector_index+1) {
         msg(MSG_ERR, "Cyl %d head %d Missed sector between %d(%d) and %d(%d)\n",
               sector_status->cyl, sector_status->head,
               drive_params->sector_numbers[orig_sector_index], orig_sector_index,
               drive_params->sector_numbers[*sector_index], *sector_index);
      }
      if (*sector_index >= drive_params->num_sectors) {
         msg(MSG_ERR_SERIOUS,"Cyl %d head %d Sector %d not found in expected sector list after %d(%d)\n",
               sector_status->cyl, sector_status->head,
               sector_status->sector,
               drive_params->sector_numbers[orig_sector_index],
               orig_sector_index);
         sector_status->status |= SECT_BAD_HEADER;
         *sector_index = orig_sector_index;
      }
      sector_status->logical_sector = *sector_index;
   } else {
      sector_status->logical_sector = *sector_index;
      (*sector_index)++;
   }
   if (sector_size != drive_params->sector_size) {
      msg(MSG_ERR,"Expected sector size %d header size %d cyl %d head %d sector %d\n",
            drive_params->sector_size, sector_size, sector_status->cyl,
            sector_status->head, sector_status->sector);
   }
}

// Write the sector data to file. We only write the best data so if the
// caller retries read with error we won't overwrite good data if this
// read has an error for this sector but the previous didn't.
// Bytes is data to write.
// drive_params: specifies the length and other information.
// sector_status: is the status of the sector writing
// sector_status_list: is the status of the data that would be written to the
//    file. Status is still updated if decode_fd is -1 to prevent writing.
// return: -1 if error found, 0 if OK.
int mfm_write_sector(uint8_t bytes[], DRIVE_PARAMS * drive_params,
      SECTOR_STATUS *sector_status,
      SECTOR_STATUS sector_status_list[])
{
   int rc;
   STATS *stats = &drive_params->stats;

   // Some disks number sectors starting from 1. We need them starting
   // from 0.
   int sect_rel0 = sector_status->sector - drive_params->first_sector_number;

   // Collect statistics
   stats->max_sect = MAX(sector_status->sector, stats->max_sect);
   stats->min_sect = MIN(sector_status->sector, stats->min_sect);
   stats->max_head = MAX(sector_status->head, stats->max_head);
   stats->min_head = MIN(sector_status->head, stats->min_head);
   stats->max_cyl = MAX(sector_status->cyl, stats->max_cyl);
   stats->min_cyl = MIN(sector_status->cyl, stats->min_cyl);

   // Check for sector and head in range to prevent bad writes.
   // We don't check cyl against max since if it exceeds it things
   // will still work properly.
   if (sect_rel0 >= drive_params->num_sectors || sect_rel0 < 0) {
      msg(MSG_ERR_SERIOUS, "Logical sector %d out of range 0-%d sector %d cyl %d head %d\n",
            sect_rel0, drive_params->num_sectors, sector_status->sector,
            sector_status->cyl,sector_status->head);
      return -1;
   }
   if (sector_status->head > drive_params->num_head) {
      msg(MSG_ERR_SERIOUS,"Head out of range %d max %d cyl %d sector %d\n", 
            sector_status->head, drive_params->num_head, 
            sector_status->cyl, sector_status->sector);
      return -1;
   }

   // If not written then write data. Otherwise only write if likely to be
   // better than sector previously written. Better is if we didn't get a CRC
   // error, or the ECC correction span is less than the last one.
   // We assume the header data is correct so we don't check if the header
   // ECC correction is better. If the header wasn't right we wrote the
   // data to the wrong spot in the file.
   if ((sector_status_list[sect_rel0].status & SECT_BAD_HEADER) ||
         (!(sector_status->status & SECT_BAD_DATA) &&
               sector_status->ecc_span_corrected_data == 0) ||
         (sector_status->ecc_span_corrected_data > 0 &&
               sector_status->ecc_span_corrected_data <
               sector_status_list[sect_rel0].ecc_span_corrected_data)) {
      if (decode_fd >= 0) {
         if (lseek(decode_fd,
               (sect_rel0) * drive_params->sector_size +
               sector_status->head * (drive_params->sector_size *
                     drive_params->num_sectors) +
                     (off_t) sector_status->cyl * (drive_params->sector_size *
                           drive_params->num_sectors *
                           drive_params->num_head), SEEK_SET) < 0) {
            msg(MSG_FATAL, "Seek failed decoded data: %s\n", strerror(errno));
            exit(1);
         };
         if ((rc = write(decode_fd, bytes, drive_params->sector_size)) !=
               drive_params->sector_size) {
            msg(MSG_FATAL, "Write failed, rc %d: %s", rc, strerror(errno));
            exit(1);
         }
      }
      sector_status_list[sect_rel0] = *sector_status;
   }
   //FIXME DELETE. This was for checking this read against previous known good
   // data
#if 0
   if (!(sector_status->status & SECT_BAD_DATA)) {
      static int fd = -1;
      uint8_t buf[512];
      int i, rc;
      if (fd == -1)
         fd = open("/tmp/decoded_out.good",O_RDONLY);
      rc = lseek(fd,
            (sect_rel0) * drive_params->sector_size +
            sector_status->head * (drive_params->sector_size *
                  drive_params->num_sectors) +
                  (off_t) sector_status->cyl * (drive_params->sector_size *
                        drive_params->num_sectors *
                        drive_params->num_head), SEEK_SET);
      read(fd, buf, 512);
      //printf("Checking %d head %d sect %d %d\n", cyl, head, sect, rc);
      for (i = 0; i < 512; i++) {
         //printf("%02x %02x\n", buf[i], bytes[i]);
         if (buf[i] != bytes[i])
            break;
      }
      if (i != 512)
         printf("Mismatch cyl %d head %d sect %d %d\n", sector_status->cyl, sector_status->head,
               sector_status->sector, i);
   }
#endif
   return 0;
}

// This prints the header or data bytes for decoding new formats
void mfm_dump_bytes(uint8_t bytes[], int len, int cyl, int head, int sector_index)
{
   int i;

   msg(MSG_DEBUG, "%4d %2d %2d:", cyl, head, sector_index);
   for (i = 0; i < len+4; i++) {
      //      MSG(MSG_INFO, " %02x",bytes[i]);
      msg(MSG_DEBUG, "0x%02x,",bytes[i]);
      if (i % 16 == 15) {
         msg(MSG_DEBUG, "\n");
      }
   }
   msg(MSG_DEBUG, "\n");
}


// After we have found a valid header/data mark this routine is
// used to process the bytes. It checks the CRC and does ECC if needed then
// calls wd_process_data to finish the processing.
//
// drive_params: Drive parameters
// bytes: bytes to process
// bytes_crc_len: Length of bytes including CRC
// cyl,head: Physical Track data from
// sector_index: Sequential sector counter
// seek_difference: Return of difference between expected cyl and header
// sector_status_list: Return of status of decoded sector
// return: Status of sector decoded
int mfm_process_bytes(DRIVE_PARAMS *drive_params, uint8_t bytes[],
      int bytes_crc_len, STATE_TYPE *state, int cyl, int head, int *sector_index,
      int *seek_difference, SECTOR_STATUS sector_status_list[]) {
   uint64_t crc;
   // CRC to use to process these bytes
   CRC_INFO crc_info;
   char *name;
   // Length of ECC correction. 0 is no correction.
   int ecc_span = 0;
   SECTOR_DECODE_STATUS status = SECT_NO_STATUS;

   if (*state == PROCESS_HEADER) {
      crc_info = drive_params->header_crc;
      if (msg_get_err_mask() & MSG_DEBUG) {
         mfm_dump_bytes(bytes, bytes_crc_len, cyl, head, *sector_index);
      }
      name = "header";
   } else {
      crc_info = drive_params->data_crc;
      if (msg_get_err_mask() & MSG_DEBUG) {
         mfm_dump_bytes(bytes, bytes_crc_len, cyl, head, *sector_index);
      }
      name = "data";
   }
   crc = crc64(bytes, bytes_crc_len, &crc_info);
   // Zero CRC is no error
   if (crc == 0) {
      status |= SECT_ZERO_CRC;
   }

   if (crc != 0) {
      //msg(MSG_ERR,"Bad CRC %s cyl %d head %d\n", name, cyl, head);
      // If ECC correction enabled then perform correction up to length
      // specified
      if (crc_info.ecc_max_span != 0) {
         ecc_span = ecc64(bytes, bytes_crc_len, crc, &crc_info);
         if (ecc_span != 0) {
            drive_params->stats.max_ecc_span = MAX(ecc_span,
                  drive_params->stats.max_ecc_span);
            crc = 0; // No longer have CRC error
         }
      }
   }

   // If no error process. Only process with errors if data. Without
   // valid header we don't know what sector we are decoding.
   if (*state != PROCESS_HEADER || crc == 0 || ecc_span != 0) {

      // If this is changed change in mfm_decode_track also
      if (drive_params->controller == CONTROLLER_WD_1006 ||
            drive_params->controller == CONTROLLER_OMTI_5510 ||
            drive_params->controller == CONTROLLER_DEC_RQDX3) {
         status |= wd_process_data(state, bytes, crc, cyl, head, sector_index,
               drive_params, seek_difference, sector_status_list, ecc_span);
      } else {
         status |= xebec_process_data(state, bytes, crc, cyl, head, sector_index,
               drive_params, seek_difference, sector_status_list, ecc_span);
      }
   } else {
      status |= SECT_BAD_HEADER;
      // Search for header in case we are out of sync. If we found
      // data next we can't process it anyway.
      *state = MARK_ID;
   }
   return status;
}
