// This is a utility program to process existing MFM delta transition data.
// Used to extract the sector contents to a file
//
// deltas_get_count routine is used by called MFM decoding routines
//
// Copyright 2013 David Gesswein.
// This file is part of MFM disk utilities.
//
// MFM disk utilities is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// MFM disk utilities is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with MFM disk utilities.  If not, see <http://www.gnu.org/licenses/>.
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <stdint.h>
#include <errno.h>

#include "msg.h"
#include "crc_ecc.h"
#define DEF_DATA
#include "mfm_decoder.h"
#include "parse_cmdline.h"
// We don't use the versions from deltas_read.h since we aren't using beaglebone PRU
// we include it to make sure our versions at the end of the file match the
// parameters the other routines are using.
#include "deltas_read.h"

#define MAX_DELTAS 131072
// The number of deltas we have to process
uint32_t num_deltas;

static void deltas_update_count(int num_deltas_in);

// Read the data for a track into deltas.
//
// deltas: Delta transition data returned
// max_deltas: Size of deltas buffer in words
// cyl, head: Track deltas are for
// fn: Filename to read data from
// return: Number of delta words read
static int read_track(uint16_t deltas[], int max_deltas, int *cyl, int *head,
      char *fn)
{
   static int fd = -1;
   int bytes;
   static uint32_t fileid;
   int rc;

   if (fd == -1) {
      fd = open(fn,O_RDONLY);
      if (fd < 0) {
         msg(MSG_FATAL,"open failed %s\n",strerror(errno));
         exit(1);
      }
      if (read(fd, &fileid, sizeof(fileid)) < sizeof(fileid)) {
         msg(MSG_FATAL,"Short read fileid\n");
         exit(1);
      }
      if (fileid != 1) {
         msg(MSG_FATAL,"Unexpected File ID word %d expected 1\n");
         exit(1);
      }
   }

   if ((rc = read(fd, &bytes, sizeof(bytes))) < sizeof(bytes)) {
      if (rc == 0) {
         return -1;
      } else {
         msg(MSG_FATAL,"Short read\n");
         exit(1);
      }
   } else {
      if (bytes / sizeof(deltas[0]) >= max_deltas) {
         msg(MSG_FATAL,"Too many deltas\n");
         exit(1);
      }
      if (read(fd, cyl, sizeof(*cyl)) < sizeof(*cyl)) {
         msg(MSG_FATAL,"Short read\n");
         exit(1);
      }
      if (read(fd, head, sizeof(*head)) < sizeof(*head)) {
         msg(MSG_FATAL,"Short read\n");
         exit(1);
      }
      if (read(fd, deltas, bytes) < bytes) {
         msg(MSG_FATAL,"Short read\n");
         exit(1);
      }
      return bytes / sizeof(deltas[0]);
   }
}

// Main routine
int main (int argc, char *argv[])
{
   uint16_t deltas[MAX_DELTAS];
   int num_deltas;
   int cyl, head;
   int last_cyl = -1, last_head = -1;
   DRIVE_PARAMS drive_params;
   SECTOR_STATUS sector_status_list[MAX_SECTORS];
   int seek_difference;
   SECTOR_DECODE_STATUS status;
   int analyze;

   // Get what to do. Turn off options we don't support
   parse_cmdline(argc, argv, &drive_params, &analyze, "arj");

   if (drive_params.transitions_filename == NULL) {
      msg(MSG_FATAL, "Transitions filename must be specified\n");
      exit(1);
   }

   // Setup decoding of transitions and possible file to write to
   mfm_decode_setup(drive_params.extract_filename, &drive_params);

   // Read and process a track at a time until all read
   while ((num_deltas = read_track(deltas, MAX_DELTAS, &cyl, &head,
         drive_params.transitions_filename)) > 0) {
      // Only clear status if we are moving to the next track. If retries
      // were done we may have multiple reads of the same track.
      if (last_cyl != cyl || last_head != head) {
         mfm_init_sector_status_list(sector_status_list, drive_params.num_sectors);
      }
      //printf("Decoding new track %d %d\n",cyl, head);
      deltas_update_count(num_deltas);
      // If head & cylinder haven't changed assume it's a retry.
      status = mfm_decode_track(&drive_params, cyl, head, deltas, &seek_difference,
            sector_status_list);
#if 0
      if (1 || status != SECT_HEADER_FOUND) {
         if (cyl == 0 && head == 0) {
            FILE *file;
            char fn[256];
            int i;
            sprintf(fn, "trk%d-%d.txt",cyl, head);
            file = fopen(fn, "w");
            for (i = 0; i < num_deltas; i++) {
               fprintf(file, "%d\n",deltas[i]);
            }
            fclose(file);
         }
      }
#endif
      last_cyl = cyl;
      last_head = head;
   }
   mfm_decode_done(&drive_params);
   return 0;
}

// Update the count of deltas.
//
// num_deltas_in: Number of deltas read
static void deltas_update_count(int num_deltas_in)
{
   num_deltas = num_deltas_in;
}

// This routines replace the routines from delta=_read that are used
// when reading from a real drive.

// Return deltas available or -1 if caller has processed all the deltas.
//
// cur_delta: Number of deltas processed so far
// return: Number of deltas available or -1 if all processed.
int deltas_get_count(int deltas_processed)
{
   if (deltas_processed >= num_deltas) {
      return -1;
   } else {
      return num_deltas;
   }
}
