// Routines for interfacing with the PRU/PRUSS. Some of the routines are
// generic and others specific to communicating with the code running on
// the PRU. Only PRU 0 is currently used.
//
// Call pru_setup first to open PRU device and map memory.
// Call pru_shutdown when done to close PRU device.
// These routines are specific to the code running on the PRU.
// Call pru_exec_cmd to send a command to PRU and wait for response.
// Call pru_get_cmd_status to get the command completion status.
// Call pru_get_cmd_data to get data value set by a command.
// Call pru_get_drive_status to get the drive status.
// Call pru_print_drive_status to decode and print the drive status.
// Call pru_restart to restart PRU from the restart address.
// Call pru_copy_deltas to copy delta data from shared DDR memory.
//
// TODO: Use cache control to make memory transfers faster with PRU
//
// Copyright 2014 David Gesswein.
// This file is part of MFM disk utilities.
//
// MFM disk utilities is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// MFM disk utilities is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with MFM disk utilities.  If not, see <http://www.gnu.org/licenses/>.
#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <unistd.h>
#include <string.h>

#include <prussdrv.h>
#include <pruss_intc_mapping.h>

#include "msg.h"
// INDD is used to convert the cmd.h defines to word indexes
#define INDD 4
#include "cmd.h"

#include "crc_ecc.h"
#include "mfm_decoder.h"
#include "drive.h"
#include "pru_setup.h"
#define ARRAYSIZE(x) (sizeof(x) / sizeof(x[0]))

// This is memory for communicating with the PRU/PRUSS processors
// DDR is the shared region of main memory, pru_dataram# is data memory for pru #.
// pru_sharedram is the shared PRU local memory
static volatile uint32_t *ddr_mem, *pru_dataram0, *pru_dataram1, *pru_sharedram;
static uint32_t ddr_phys_addr;
static int ddr_mem_size;

// Map DDR shared memory segment into our address space and return addresses
// and size.
//
// ddrmem: Returns pointer to memory (virtual address).
// ddr_phys_addr: Returns physical address of memory.
// return: Size of region in bytes.
static int pru_allocate_ddr_memory(volatile uint32_t **ddrmem, uint32_t *ddr_phys_addr)
{
   uint32_t ddr_offset, ddr_mem_size;
   FILE *fin;

   fin = fopen("/sys/class/uio/uio0/maps/map1/addr", "r");
   if (fin == NULL) {
      perror("Unable to open DDR map address");
      exit(1);
   }
   fscanf(fin, "%x", ddr_phys_addr);
   fclose(fin);

   fin = fopen("/sys/class/uio/uio0/maps/map1/offset", "r");
   if (fin == NULL) {
      perror("Unable to open DDR map offset");
      exit(1);
   }
   fscanf(fin, "%x", &ddr_offset);
   fclose(fin);

   fin = fopen("/sys/class/uio/uio0/maps/map1/size", "r");
   if (fin == NULL) {
      perror("Unable to open DDR map address");
      exit(1);
   }
   fscanf(fin, "%x", &ddr_mem_size);
   fclose(fin);
   //printf("DDR base %x offset %x size %x\n", ddr_base, ddr_offset,
   //       ddrMemSize);

   *ddrmem = prussdrv_get_virt_addr(*ddr_phys_addr + ddr_offset);

   return (ddr_mem_size);
}

// Set up PRU device and map memory
//
// return: size of DDR memory segment in bytes.
int  pru_setup()
{
   tpruss_intc_initdata pruss_intc_initdata = PRUSS_INTC_INITDATA;
   int ret;

   /* Initialize the PRU */
   prussdrv_init();

   /* Open PRU Interrupt */
   ret = prussdrv_open(PRU_EVTOUT_0);
   if (ret) {
      msg(MSG_FATAL, "prussdrv_open open failed\n");
      return -1;
   }

   /* Get the interrupt initialized */
   prussdrv_pruintc_init(&pruss_intc_initdata);

   /* Allocate PRU memory. */
   if (prussdrv_map_prumem(PRUSS0_PRU0_DATARAM,(void **) &pru_dataram0) == -1) {
      msg(MSG_FATAL, "prussdrv_map_prumem PRUSS0_PRU0_DATARAM map failed\n");
      return -1;
   }
   if (prussdrv_map_prumem(PRUSS0_PRU1_DATARAM,(void **) &pru_dataram1) == -1) {
      msg(MSG_FATAL, "prussdrv_map_prumem PRUSS0_PRU1_DATARAM map failed\n");
      return -1;
   }
   if (prussdrv_map_prumem(PRUSS0_SHARED_DATARAM,(void **) &pru_sharedram) == -1) {
      msg(MSG_FATAL, "prussdrv_map_prumem PRUSS0_SHARED_DATARAM map failed\n");
      return -1;
   }

   ddr_mem_size = pru_allocate_ddr_memory(&ddr_mem, &ddr_phys_addr);

   // Give the DDR memory region address to the PRU
   pru_dataram0[PRU0_DDR_ADDR] = ddr_phys_addr;
   pru_dataram0[PRU0_DDR_SIZE] = ddr_mem_size - 1;

   return ddr_mem_size;
}

// Wait for completion from PRU and shut down device.
void pru_shutdown() {
   // Wait until PRU0 has finished execution
   prussdrv_pru_wait_event (PRU_EVTOUT_0);
   prussdrv_pru_clear_event (PRU0_ARM_INTERRUPT, PRU_EVTOUT_0);
   prussdrv_pru_disable(0);
   //   prussdrv_pru_disable(1);
   prussdrv_exit ();
}


// This sends a command and data word to the PRU and waits for completion or,
// for reads, that the read has started. To execute a command first write the
// data value if needed then write the command to the command address. The PRU
// will execute the command and update the command address with the command
// status. The read command which we wish to overlap processing data with
// reading returns a command started status and then a command done status.
// See cmd.h for commands.
//
// cmd: Command to execute.
// data: Data value for command.
int pru_exec_cmd(uint32_t cmd, uint32_t data)
{
   // The command location is also used for status. Wait for it to change
   // from what we write
   pru_dataram0[PRU0_CMD_DATA] = data;
   pru_dataram0[PRU0_CMD] = cmd;
   while (pru_dataram0[PRU0_CMD] == cmd) {
      usleep(500);
   };
   // If not an expected result print an error message
   if (pru_dataram0[PRU0_CMD] != CMD_STATUS_OK &&
         pru_dataram0[PRU0_CMD] != CMD_STATUS_READ_STARTED) {
      msg(MSG_ERR, "Command %d fault %x status %x\n", cmd, pru_dataram0[PRU0_CMD],
            pru_dataram0[PRU0_DRIVE_STATUS]);
      pru_print_drive_status(MSG_ERR, pru_dataram0[PRU0_DRIVE_STATUS]);
      return pru_dataram0[PRU0_CMD];
   }

   return 0;
}

// Return the drive status value. See print routine for bit definitions.
uint32_t pru_get_drive_status() {
   return pru_dataram0[PRU0_DRIVE_STATUS];
}

// Return the command status value. See cmd.h for values.
uint32_t pru_get_cmd_status() {
   return pru_dataram0[PRU0_CMD];
}

// Return the command data value. Not all commands return data.
uint32_t pru_get_cmd_data() {
   return pru_dataram0[PRU0_CMD_DATA];
}

// Return the DDR write pointer. Used for debugging
uint32_t pru_get_deltas_write_ptr() {
   return pru_dataram0[PRU0_WRITE_PTR];
}

// This routine copies the deltas from the PRU shared DDR memory

// Decodes and prints the disk status value from the PRU.
//
// status: Disk status value (from pru_get_drive_status)
void pru_print_drive_status(int level, uint32_t status)
{
   struct {
      char *desc;
      int bit;
   } reg_bits[] = {
         {"Write fault", R31_WRITE_FAULT_BIT},
         {"Seek complete", R31_SEEK_COMPLETE_BIT},
         {"Index", R31_INDEX_BIT},
         {"Ready", R31_READY_BIT},
         {"Drive selected", R31_DRIVE_SEL}
   };
   int n;

   for (n = 0; n < ARRAYSIZE(reg_bits); n++) {
      // Control lines are active low
      if (status & (1 << reg_bits[n].bit)) {
         msg(level, "Not %s \n", reg_bits[n].desc);
      } else {
         msg(level, "%s \n", reg_bits[n].desc);
      }
   }
   if (drive_at_track0()) {
      msg(level, "Track 0\n");
   } else {
      msg(level, "Not Track 0\n");
   }
}

// Call this routine to restart PRU0 at the restart address
void pru_restart(int pru_num)
{
   if (pru_num == 0) {
      // Set the starting address to STOP_ADDR and restart the PRU.
      pru_dataram0[PRU_CONTROL_REG] = (RESTART_ADDR << 16) | 2;
   } else {
      // Set the starting address to STOP_ADDR and restart the PRU.
      pru_dataram1[PRU_CONTROL_REG] = (RESTART_ADDR << 16) | 2;
   }
}

// to local memory. The DDR memory is uncached so slow.
// TODO: Should use cached memory and use cache flush/invalidate
// to improve performance.
// Internal to the routine we use bytes. The caller uses word
// indexes so we need to convert by 4. Buffer is linear and
// reused for each track so we don't have to deal with wrapping.
int pru_copy_deltas(uint16_t deltas[],int last_index)
{
   int bytes;
   uint16_t *ddr_mem_16 = (uint16_t *) ddr_mem;

   bytes = pru_dataram0[PRU0_WRITE_PTR] - last_index * sizeof(deltas[0]);
   if (bytes > ddr_mem_size || bytes < 0) {
      msg(MSG_FATAL, "Read too many transistions %d %d %d\n", bytes, last_index,
            pru_dataram0[PRU0_WRITE_PTR]);
      exit(1);
   }
   memcpy(&deltas[last_index], (unsigned int *) &ddr_mem_16[last_index], bytes);
   return bytes / sizeof(deltas[0]);
}
